/*
 * Fabric3
 * Copyright  2008 Metaform Systems Limited
 *
 * This proprietary software may be used only connection with the Fabric3 license
 * (the License), a copy of which is included in the software or may be
 * obtained at: http://www.metaformsystems.com/licenses/license.html.

 * Software distributed under the License is distributed on an as is basis,
 * without warranties or conditions of any kind.  See the License for the
 * specific language governing permissions and limitations of use of the software.
 * This software is distributed in conjunction with other software licensed under
 * different terms.  See the separate licenses for those programs included in the
 * distribution for the permitted and restricted uses of such software.
 *
 */
package org.fabric3.fabric.services.archive;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.RandomAccessFile;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.osoa.sca.annotations.EagerInit;
import org.osoa.sca.annotations.Init;
import org.osoa.sca.annotations.Reference;

import org.fabric3.host.RuntimeMode;
import org.fabric3.host.runtime.HostInfo;
import org.fabric3.spi.services.archive.ArchiveStore;
import org.fabric3.spi.services.archive.ArchiveStoreException;

/**
 * The default implementation of ArchiveStore
 *
 * @version $Rev: 6590 $ $Date: 2009-02-21 12:24:13 -0800 (Sat, 21 Feb 2009) $
 */
@EagerInit
public class ArchiveStoreImpl implements ArchiveStore {
    private Map<URI, URL> archiveUriToUrl;
    private File repository;
    private boolean participant;

    /**
     * Constructor.
     *
     * @param hostInfo the host info for the runtime
     * @throws IOException if an error occurs initializing the repository
     */
    public ArchiveStoreImpl(@Reference HostInfo hostInfo) throws IOException {
        archiveUriToUrl = new ConcurrentHashMap<URI, URL>();
        File baseDir = hostInfo.getBaseDir();
        repository = new File(baseDir, "repository");
        participant = hostInfo.getRuntimeMode() == RuntimeMode.PARTICIPANT;
    }

    @Init
    public void init() throws MalformedURLException {
        if (!repository.exists() || !repository.isDirectory()) {
            return;
        }
        if (participant) {
            // if the runtime is in participant mode, clear out cached archives as different contributions may be provisioned on restart
            for (File file : repository.listFiles()) {
                file.delete();
            }
        } else {
            // load archives
            for (File file : repository.listFiles()) {
                archiveUriToUrl.put(mapToUri(file), file.toURI().toURL());
            }
        }
    }

    public URL store(URI uri, InputStream stream) throws ArchiveStoreException {
        try {
            if (!repository.exists() || !repository.isDirectory() || !repository.canRead()) {
                throw new IOException("The repository location is not a directory: " + repository);
            }
            File location = mapToFile(uri);
            write(stream, location);
            URL locationUrl = location.toURL();
            archiveUriToUrl.put(uri, locationUrl);
            return locationUrl;
        } catch (IOException e) {
            String id = uri.toString();
            throw new ArchiveStoreException("Error storing archive: " + id, id, e);
        }
    }

    public boolean exists(URI uri) {
        return archiveUriToUrl.containsKey(uri);
    }

    public URL find(URI uri) {
        return archiveUriToUrl.get(uri);
    }

    public void remove(URI uri) throws ArchiveStoreException {
        try {
            File location = mapToFile(uri);
            archiveUriToUrl.remove(uri);
            location.delete();
        } catch (IOException e) {
            String id = uri.toString();
            throw new ArchiveStoreException("Error storing archive: " + id, id, e);
        }
    }

    public List<URI> list() {
        return new ArrayList<URI>(archiveUriToUrl.keySet());
    }

    /**
     * Resolve contribution location in the repository.
     *
     * @param uri the uri to resolve
     * @return the mapped file
     * @throws IOException if an exception occurs mapping the file
     */
    private File mapToFile(URI uri) throws IOException {
        if (!repository.exists() || !repository.isDirectory() || !repository.canRead()) {
            throw new IOException("The repository location is not a directory: " + repository);
        }
        return new File(repository, uri.getPath());
    }

    /**
     * Maps a file to a URI.
     *
     * @param file the file
     * @return the URI
     */
    private URI mapToUri(File file) {
        return URI.create(file.getName());
    }

    private void write(InputStream source, File target) throws IOException {
        RandomAccessFile file = new RandomAccessFile(target, "rw");
        FileChannel channel = null;
        FileLock lock = null;
        try {
            channel = file.getChannel();
            lock = channel.lock();
            ByteBuffer buffer = ByteBuffer.allocate(1024);
            byte[] bytes = buffer.array();
            int limit;
            while (-1 != (limit = source.read(bytes))) {
                buffer.flip();
                buffer.limit(limit);
                while (buffer.hasRemaining()) {
                    channel.write(buffer);
                }
                buffer.clear();
            }
            channel.force(true);
        } finally {
            if (channel != null) {
                if (lock != null) {
                    lock.release();
                }
                channel.close();
            }
            file.close();
        }

    }
}
